import time
import traceback
import uuid

import attr
import boto3
import botocore

from pyflow import decision_task_helper as dth
from pyflow import exceptions
from pyflow import utils
from pyflow import workflow_invocation_helper as wih
from pyflow import workflow_state as ws
from pyflow.utils import logger

__all__ = ['Decider', 'get_workflow_status', 'poll_for_executions', 'start_workflow', 'WorkflowStatus']


dashboard_url_template = 'https://{region}.console.aws.amazon.com/swf/home?region={region}#' \
                         'execution_summary:domain={domain};workflowId={workflow_id};runId={run_id}'


class Decider(object):
    def __init__(self, workflow_classes, domain, task_list, identity, client=None):
        self._workflow_classes = {(w.NAME, w.VERSION): w for w in workflow_classes}

        self._domain = domain
        self._task_list = task_list
        self._identity = identity

        if client is None:
            # Read timeout needs to be > 60 seconds, because the poll_for_decision_tasks API call waits up to 60
            # seconds before returning.
            client = boto3.client('swf', config=botocore.client.Config(read_timeout=70))

        self._client = client

    def ensure_workflows_registered(self):
        """
        Register all the workflow types passed to the constructor with SWF, if they are not already registered
        """
        for workflow_class in self._workflow_classes.values():
            workflow_obj = workflow_class(None)
            workflow_type = dict(name=workflow_obj.name, version=workflow_obj.version)
            options = {k: str(v) for k, v in workflow_obj.options.items()}

            try:
                self._client.register_workflow_type(
                    domain=self._domain,
                    name=workflow_type['name'],
                    version=workflow_type['version'],
                    defaultTaskList={'name': self._task_list},
                    **options)
            except self._client.exceptions.TypeAlreadyExistsFault:
                logger.debug('Workflow %r already registered', workflow_type)
            else:
                logger.info("Registered workflow: %r", workflow_type)

    def process_decision_task(self, decision_task):
        """
        Handle a single decision task.  Takes care of updating the workflow state, and computing decisions, but doesn't
        actually send the response to SWF.

        :param decision_task: An SWF DecisionTask object
        """

        workflow_id = decision_task['workflowExecution']['workflowId']
        run_id = decision_task['workflowExecution']['runId']

        workflow_state = ws.WorkflowState(workflow_id=workflow_id, run_id=run_id)

        decision_helper = dth.DecisionTaskHelper(decision_task, workflow_state)

        if decision_helper.previous_started_event_id == 0:
            dashboard_url = dashboard_url_template.format(region=boto3.Session().region_name, domain=self._domain,
                                                          workflow_id=workflow_id, run_id=run_id)
            logger.info('Starting instance of workflow {}: {}'.format(
                {'name': decision_helper.workflow_name, 'version': decision_helper.workflow_version},
                {'workflow_id': decision_helper.workflow_id, 'run_id': decision_helper.run_id}))
            logger.info('Dashboard URL: {}'.format(dashboard_url))

        try:
            workflow_type = (decision_helper.workflow_name, decision_helper.workflow_version)
            workflow_class = self._workflow_classes.get(workflow_type)
            if workflow_class is None:
                raise exceptions.DeciderException('Received decision task for unknown workflow type: {!r}'.format(
                    workflow_type))

            # Process the first decision task's events to fill in workflow input
            decision_helper.process_next_decision_task()

            invocation_helper = wih.WorkflowInvocationHelper(decision_helper)
            workflow_obj = workflow_class(invocation_helper)
            result = workflow_obj.run(decision_helper.workflow_state.input)
        except exceptions.WorkflowBlockedException:
            pass
        except exceptions.WorkflowFailedException as e:
            logger.info('Workflow function threw WorkflowFailedException(reason={!r}, details={!r})'.format(
                e.reason, e.details))
            decision_helper.fail_workflow(e.reason, e.details)
        except Exception:
            msg = 'Caught exception while executing workflow {!r}'.format(workflow_type)
            logger.exception(msg)
            decision_helper.fail_workflow(msg, traceback.format_exc())
        else:
            try:
                decision_helper.complete_workflow(result)
            except Exception:
                msg = 'Caught exception while processing workflow result for workflow {!r}'.format(workflow_type)
                logger.exception(msg)
                decision_helper.fail_workflow(msg, traceback.format_exc())

        return decision_helper

    def poll_for_decision_tasks(self, max_time=None):
        self.ensure_workflows_registered()

        start_time = time.time()
        logger.info('Beginning to poll for decisions in domain {!r}, task_list {!r}'.format(
            self._domain, self._task_list))

        # save_dir = '/tmp/decision_tasks'
        # if not os.path.exists(save_dir):
        #     os.makedirs(save_dir)
        # save_counter = 0

        while max_time is None or time.time() < (start_time + max_time):
            decision_task = utils.poll_for_decision_tasks(
                self._client, self._domain, self._task_list, self._identity)
            if not decision_task.get('taskToken'):
                # poll timed out
                logger.debug('Poll timed out')
                continue

            logger.debug('Processing decision task')

            # with open(os.path.join(save_dir, 'decision_task{:02}.pickle'.format(save_counter)), 'w') as f:
            #     cPickle.dump(decision_task, f)
            #     save_counter += 1

            decision_helper = self.process_decision_task(decision_task)

            try:
                self._client.respond_decision_task_completed(
                    taskToken=decision_helper.task_token,
                    decisions=decision_helper.decisions)
            except Exception:
                logger.exception('respond_decision_task_completed failed')

                # Assume it failed due to a bad decision object.  Clear decisions list and fail the workflow
                decision_helper.decisions = []
                decision_helper.fail_workflow('respond_decision_task_completed failed', traceback.format_exc())
                self._client.respond_decision_task_completed(
                    taskToken=decision_helper.task_token,
                    decisions=decision_helper.decisions)


def poll_for_executions(workflows, domain, task_list, identity, max_time=None):
    the_decider = Decider(workflows, domain, task_list, identity)
    the_decider.poll_for_decision_tasks(max_time)


def start_workflow(domain, workflow_name, workflow_version, task_list, lambda_role, input=None, client=None):
    workflow_id = '{}@{}'.format(workflow_name, str(uuid.uuid4()))

    if client is None:
        client = boto3.client('swf')

    response = client.start_workflow_execution(
        domain=domain,
        workflowId=workflow_id,
        workflowType={'name': workflow_name, 'version': workflow_version},
        taskList={'name': task_list},
        input=input,
        lambdaRole=lambda_role)

    return {'workflowId': workflow_id, 'runId': response['runId']}


@attr.s
class WorkflowStatus(object):
    # 'OPEN'|'CLOSED'
    execution_status = attr.ib()

    # The full response returned from DescribeWorkflowExecution. In most cases the fields below should provide the
    # information you need, but this object has the full details.
    workflow_description = attr.ib()

    # Only set if execution_status is CLOSED
    # 'COMPLETED'|'FAILED'|'CANCELED'|'TERMINATED'|'CONTINUED_AS_NEW'|'TIMED_OUT'
    close_status = attr.ib(default=None)

    # timestamp of the latest activity for this workflow
    latest_activity_timestamp = attr.ib(default=None)

    # The result of the workflow, as a string, if workflow status is CLOSED and close_status is COMPLETED
    result = attr.ib(default=None)

    # The failure reason and details strings, if close_status is FAILED
    failure_reason = attr.ib(default=None)
    failure_details = attr.ib(default=None)

    # If execution_status is CLOSED, then this will point to the event describing why the workflow exited.  Normally
    # information you need can be gotten from other fields.
    exit_event = attr.ib(default=None)


def get_workflow_status(domain, workflow_id, run_id, client=None):
    """
    Get the status of a currently running or closed workflow instance.

    :param domain: SWF Domain
    :param workflow_id: Workflow ID of the workflow instance
    :param run_id: Run ID of the workflow instance
    :param client: Boto3 SWF client
    :return: A WorkflowStatus object, or None if no workflow is found matching the workflow id and run id
    """

    if client is None:
        client = boto3.client('swf')

    try:
        description = client.describe_workflow_execution(
            domain=domain, execution={'workflowId': workflow_id, 'runId': run_id})
    except client.exceptions.UnknownResourceFault:
        return None

    exec_info = description['executionInfo']
    exec_status = exec_info['executionStatus']

    status = WorkflowStatus(execution_status=exec_status, workflow_description=description,
                            latest_activity_timestamp=description.get('latestActivityTaskTimestamp'))

    exit_event_types = {'WorkflowExecutionCompleted', 'WorkflowExecutionFailed', 'WorkflowExecutionCanceled',
                        'WorkflowExecutionTerminated', 'WorkflowExecutionContinuedAsNew', 'WorkflowExecutionTimedOut'}

    if exec_status == 'CLOSED':
        status.close_status = exec_info['closeStatus']

        for event in utils.list_workflow_history(client, domain, workflow_id, run_id, reverse=True):
            if event['eventType'] in exit_event_types:
                status.exit_event = event
                if event['eventType'] == 'WorkflowExecutionCompleted':
                    status.result = event['workflowExecutionCompletedEventAttributes'].get('result')
                elif event['eventType'] == 'WorkflowExecutionFailed':
                    attributes = event['workflowExecutionFailedEventAttributes']
                    status.failure_reason = attributes.get('reason')
                    status.failure_details = attributes.get('details')
                break

    return status



import attr


@attr.s
class InvocationState(object):
    """
    Encapsulates the state of an invocation, such as a lambda invocation, activity, timer, etc.
    """

    NOT_STARTED = 0
    HANDLED = 1
    STARTED = 2
    FAILED = 3
    TIMED_OUT = 4
    CANCELED = 5
    SUCCEEDED = 6

    DONE_STATES = (FAILED, TIMED_OUT, CANCELED, SUCCEEDED)

    invocation_id = attr.ib()

    state = attr.ib(default=NOT_STARTED)

    result = attr.ib(default=None)

    failure_reason = attr.ib(default=None)
    failure_details = attr.ib(default=None)
    
    @property
    def done(self):
        return self.state in self.DONE_STATES

    def update_state(self, state=None, result=None, failure_reason=None, failure_details=None):
        if state is not None:
            self.state = state
        if result is not None:
            self.result = result
        if failure_reason is not None:
            self.failure_reason = failure_reason
        if failure_details is not None:
            self.failure_details = failure_details


@attr.s
class WorkflowState(object):
    """Encapsulates the state of a workflow instance"""

    # Identifies the workflow instance
    workflow_id = attr.ib()
    run_id = attr.ib()

    # The ARN of the lambda role specified when starting this workflow
    lambda_role = attr.ib(default=None)

    # A datetime object identifying when this workflow instance started
    workflow_start_time = attr.ib(default=None)

    # Input to the workflow when it was started
    input = attr.ib(default=None)

    # True if this workflow has completed, whether successfully or not.
    completed = attr.ib(default=False)

    # A dictionary of invocation states.  Keys are invocation ids, and values are InvocationState objects.
    invocation_states = attr.ib(default=attr.Factory(dict))  # type: dict[str, InvocationState]

    # The id of the last event added to the state
    last_seen_event_id = attr.ib(default=None)

    def get_invocation_state(self, invocation_id, initial_state=InvocationState.NOT_STARTED):
        """
        Gets the invocation state for an invocation_id, creating a new state if none exists

        :param invocation_id: The invocation id of the state to fetch
        :param initial_state: The initial value to set the state property to if a new InvocationState is created
        :return: The InvocationState object for invocation_id
        """
        invocation_state = self.invocation_states.get(invocation_id)
        if invocation_state is None:
            invocation_state = self.invocation_states[invocation_id] = InvocationState(
                invocation_id=invocation_id, state=initial_state)
        return invocation_state



import re

from pyflow import utils
from pyflow import workflow_state as ws
from pyflow.utils import logger


class EventHandler(object):
    # A (almost) complete list of possible event types is available at
    # http://docs.aws.amazon.com/amazonswf/latest/apireference/API_HistoryEvent.html
    # The list is missing the lambda events

    # Categorized list of event types that can be sent to a decider
    event_types = {
        'ActivityTask': [
            'ActivityTaskCancelRequested',
            'ActivityTaskCanceled',
            'ActivityTaskCompleted',
            'ActivityTaskFailed',
            'ActivityTaskScheduled',
            'ActivityTaskStarted',
            'ActivityTaskTimedOut',
            'RequestCancelActivityTaskFailed',
            'ScheduleActivityTaskFailed',
            'StartActivityTaskFailed'],
        'Timer': [
            'CancelTimerFailed',
            'StartTimerFailed',
            'TimerCanceled',
            'TimerFired',
            'TimerStarted'],
        'WorkflowExecution': [
            'CancelWorkflowExecutionFailed',
            'CompleteWorkflowExecutionFailed',
            'FailWorkflowExecutionFailed',
            'WorkflowExecutionCancelRequested',
            'WorkflowExecutionCanceled',
            'WorkflowExecutionCompleted',
            'WorkflowExecutionFailed',
            'WorkflowExecutionSignaled',
            'WorkflowExecutionStarted',
            'WorkflowExecutionTerminated',
            'WorkflowExecutionTimedOut'],
        'ChildWorkflowExecution': [
            'ChildWorkflowExecutionCanceled',
            'ChildWorkflowExecutionCompleted',
            'ChildWorkflowExecutionFailed',
            'ChildWorkflowExecutionStarted',
            'ChildWorkflowExecutionTerminated',
            'ChildWorkflowExecutionTimedOut',
            'StartChildWorkflowExecutionFailed',
            'StartChildWorkflowExecutionInitiated'],
        'ContinueAsNewWorkflowExecution': [
            'WorkflowExecutionContinuedAsNew',
            'ContinueAsNewWorkflowExecutionFailed'],
        'DecisionTask': [
            'DecisionTaskCompleted',
            'DecisionTaskScheduled',
            'DecisionTaskStarted',
            'DecisionTaskTimedOut'],
        'ExternalWorkflowExecution': [
            'ExternalWorkflowExecutionCancelRequested',
            'ExternalWorkflowExecutionSignaled',
            'RequestCancelExternalWorkflowExecutionFailed',
            'RequestCancelExternalWorkflowExecutionInitiated',
            'SignalExternalWorkflowExecutionFailed',
            'SignalExternalWorkflowExecutionInitiated'],
        'Marker': [
            'MarkerRecorded',
            'RecordMarkerFailed'],
        'LambdaFunction': [
            'LambdaFunctionCompleted',
            'LambdaFunctionFailed',
            'LambdaFunctionScheduled',
            'LambdaFunctionStarted',
            'LambdaFunctionTimedOut',
            'ScheduleLambdaFunctionFailed',
            'StartLambdaFunctionFailed']
    }

    def __init__(self, decision_helper):
        """:type decision_helper: dth.DecisionTaskHelper"""
        self.decision_helper = decision_helper

    def update_state_from_event(self, event):
        """
        Updates the workflow state in response to a decision event

        :param event: A workflow event object
        :return: True if any invocation state changed from a not-done state to a done-state.
        """
        logger.debug('Processing event %r', event)
        self.decision_helper.workflow_state.last_seen_event_id = event['eventId']

        event_type = event['eventType']

        # handler name is computed from event type by prepending 'handle_' to the snake-case version of event type
        handler_name = 'handle_' + event_type[0].lower() + \
                       re.sub(r'([a-z])([A-Z])', lambda m: m.group(1) + '_' + m.group(2).lower(), event_type[1:])

        handler = getattr(self, handler_name, self.handle_unhandled_event)

        return handler(event)

    def _handle_state_change(self, event, new_state, failure_reason=None, failure_details=None, result=None):
        invocation_id = self.decision_helper.event_invocation_id(event)
        invocation_state = self.decision_helper.workflow_state.get_invocation_state(invocation_id)
        invocation_state.update_state(state=new_state, failure_reason=failure_reason,
                                      failure_details=failure_details, result=result)
        return invocation_state.done

    def handle_unhandled_event(self, event):
        logger.debug('Skipping handling of event %r', event['eventType'])
        return False

    # Lambda event handlers

    def handle_lambda_function_completed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        result = utils.decode_task_result(attributes.get('result'))
        return self._handle_state_change(event, ws.InvocationState.SUCCEEDED, result=result)

    def handle_lambda_function_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('reason'),
                                         failure_details=attributes.get('details'))

    def handle_lambda_function_scheduled(self, event):
        return self._handle_state_change(event, ws.InvocationState.HANDLED)

    def handle_lambda_function_started(self, event):
        return self._handle_state_change(event, ws.InvocationState.STARTED)

    def handle_lambda_function_timed_out(self, event):
        return self._handle_state_change(event, ws.InvocationState.TIMED_OUT,
                                         failure_reason='Lambda function timed out')

    def handle_start_lambda_function_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('cause'))

    def handle_schedule_lambda_function_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('cause'))

    # Workflow event handlers

    def handle_workflow_execution_started(self, event):
        attributes = self.decision_helper.event_attributes(event)
        self.decision_helper.workflow_state.workflow_start_time = event['eventTimestamp']
        self.decision_helper.workflow_state.lambda_role = attributes.get('lambdaRole')
        try:
            self.decision_helper.workflow_state.input = utils.decode_task_result(attributes.get('input'))
        except ValueError as e:
            self.decision_helper.fail_workflow('Invalid input to workflow', str(e))
        return True

    def handle_workflow_execution_timed_out(self, event):
        self.decision_helper.workflow_state.completed = True
        self.decision_helper.should_delete = True
        return False

    def handle_workflow_execution_failed(self, event):
        self.decision_helper.workflow_state.completed = True
        self.decision_helper.should_delete = True
        return False

    def handle_workflow_execution_completed(self, event):
        self.decision_helper.workflow_state.completed = True
        self.decision_helper.should_delete = True
        return False

    def handle_workflow_execution_terminated(self, event):
        self.decision_helper.workflow_state.completed = True
        self.decision_helper.should_delete = True
        return False

    def handle_workflow_execution_cancel_requested(self, event):
        attributes = self.decision_helper.event_attributes(event)
        self.decision_helper.cancel_workflow(attributes.get('cause'))
        return False

    # Timer event handlers

    def handle_start_timer_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('cause'))

    def handle_timer_fired(self, event):
        return self._handle_state_change(event, ws.InvocationState.SUCCEEDED)

    def handle_timer_started(self, event):
        return self._handle_state_change(event, ws.InvocationState.STARTED)

    def handle_timer_canceled(self, event):
        return self._handle_state_change(event, ws.InvocationState.CANCELED)

    # Activity Event Handlers

    def handle_activity_task_scheduled(self, event):
        return self._handle_state_change(event, ws.InvocationState.HANDLED)

    def handle_activity_task_started(self, event):
        return self._handle_state_change(event, ws.InvocationState.STARTED)

    def handle_activity_task_cancel_requested(self, event):
        return self.handle_unhandled_event(event)

    def handle_activity_task_completed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        try:
            result = utils.decode_task_result(attributes.get('result'))
        except ValueError as e:
            return self._handle_state_change(event, ws.InvocationState.FAILED, failure_reason='Decoding result failed',
                                             failure_details=str(e))

        return self._handle_state_change(event, ws.InvocationState.SUCCEEDED, result=result)

    def handle_activity_task_canceled(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.CANCELED,
                                         failure_details=attributes.get('details'))

    def handle_activity_task_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('reason'),
                                         failure_details=attributes.get('details'))

    def handle_activity_task_timed_out(self, event):
        return self._handle_state_change(event, ws.InvocationState.TIMED_OUT,
                                         failure_reason='Activity task timed out')

    def handle_request_cancel_activity_task_failed(self, event):
        return self.handle_unhandled_event(event)

    def handle_schedule_activity_task_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('cause'))

    def handle_start_activity_task_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('cause'))

    # Child workflow event handlers

    def handle_start_child_workflow_execution_initiated(self, event):
        return self._handle_state_change(event, ws.InvocationState.HANDLED)

    def handle_child_workflow_execution_started(self, event):
        return self._handle_state_change(event, ws.InvocationState.STARTED)

    def handle_child_workflow_execution_completed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        try:
            result = utils.decode_task_result(attributes.get('result'))
        except ValueError as e:
            return self._handle_state_change(event, ws.InvocationState.FAILED,
                                             failure_reason='Failed to decode child workflow output',
                                             failure_details=str(e))
        return self._handle_state_change(event, ws.InvocationState.SUCCEEDED, result=result)

    def handle_child_workflow_execution_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('reason'),
                                         failure_details=attributes.get('details'))

    def handle_child_workflow_execution_canceled(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.CANCELED,
                                         failure_details=attributes.get('details'))

    def handle_child_workflow_execution_terminated(self, event):
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason='Child workflow was terminated')

    def handle_child_workflow_execution_timed_out(self, event):
        return self._handle_state_change(event, ws.InvocationState.TIMED_OUT,
                                         failure_reason='Child workflow timed out')

    def handle_start_child_workflow_execution_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED,
                                         failure_reason=attributes.get('cause'))

    # marker event handlers

    def handle_marker_recorded(self, event):
        attributes = self.decision_helper.event_attributes(event)
        try:
            result = utils.decode_task_result(attributes.get('details'))
        except ValueError as e:
            return self._handle_state_change(event, ws.InvocationState.FAILED,
                                             failure_reason='Failed to decode marker value',
                                             failure_details=str(e))
        return self._handle_state_change(event, ws.InvocationState.SUCCEEDED, result=result)

    def handle_record_marker_failed(self, event):
        attributes = self.decision_helper.event_attributes(event)
        return self._handle_state_change(event, ws.InvocationState.FAILED, failure_reason=attributes['cause'])



from pyflow import event_handler as eh
from pyflow import utils
from pyflow import workflow_state as ws


class DecisionTaskHelper(object):
    def __init__(self, decision_task, workflow_state):
        """
        Initialize a DecisionTaskHelper

        :param decision_task: An SWF DecisionTask object
        :param workflow_state: WorkflowState
        :type workflow_state: ws.WorkflowState
        """
        self._decision_task = decision_task
        self._event_handler = eh.EventHandler(self)

        self.workflow_state = workflow_state
        self.decisions = []
        self.is_replaying = False
        self.last_event_timestamp = None

        # Index into the events list of the decision task of the last event processed by process_next_decision_task
        self._last_event_idx = -1

    @property
    def run_id(self):
        return self._decision_task['workflowExecution']['runId']

    @property
    def workflow_id(self):
        return self._decision_task['workflowExecution']['workflowId']

    @property
    def workflow_name(self):
        return self._decision_task['workflowType']['name']

    @property
    def workflow_version(self):
        return self._decision_task['workflowType']['version']

    @property
    def task_token(self):
        return self._decision_task['taskToken']

    @property
    def events(self):
        return self._decision_task['events']

    @property
    def previous_started_event_id(self):
        return self._decision_task['previousStartedEventId']

    @property
    def started_event_id(self):
        return self._decision_task['startedEventId']

    @property
    def decision_task(self):
        return self._decision_task

    def is_replay_event(self, event):
        return event['eventId'] <= self.previous_started_event_id

    def process_next_decision_task(self):
        """
        Process events from the current decision task until encountering a DecisionTaskStarted event or end of the
        event list.

        :return: The last event processed, or None if there are no more events to process
        """
        event = None
        for self._last_event_idx in range(self._last_event_idx + 1, len(self.events)):
            event = self.events[self._last_event_idx]
            self.is_replaying = event['eventId'] <= self.previous_started_event_id
            self.last_event_timestamp = event['eventTimestamp']
            self._event_handler.update_state_from_event(event)
            if event['eventType'] == 'DecisionTaskStarted':
                break

        return event

    def schedule_lambda_invocation(self, invocation_id, function_name, input_arg, timeout=None):
        """
        Add a ScheduleLambdaFunction decision to the list of decisions.

        :param invocation_id: The id to use for the event
        :param function_name: Lambda function name
        :param input_arg: Input argument to the lambda function.  It should be a JSON-serializable object.
        :param timeout: Timeout value, in seconds, after which the lambda function is considered to have failed
          if it has not returned.
        """
        attributes = {
            'id': invocation_id,
            'name': function_name
        }

        if input_arg is not None:
            attributes['input'] = utils.encode_task_input(input_arg)

        if timeout:
            attributes['startToCloseTimeout'] = str(timeout)

        self.decisions.append({
            'decisionType': 'ScheduleLambdaFunction',
            'scheduleLambdaFunctionDecisionAttributes': attributes
        })

    def schedule_activity_invocation(self, invocation_id, activity_name, activity_version, input_arg, timeout=None,
                                     task_list=None):
        """
        Add a ScheduleActivityTask decision to the list of decisions.

        :param invocation_id: The id to use for the event
        :param activity_name: Activity type name
        :param activity_version: Activity type version
        :param input_arg: Input argument to the activity.  It should be a JSON-serializable object.
        :param timeout: Timeout value, in seconds, after which the activity task is considered to have failed
          if it has not returned.
        :param task_list: Name of the SWF task list that the activity worker is listening for events on.  If not given,
          The default registered with the activity type will be used.
        """
        attributes = {
            'activityId': invocation_id,
            'activityType': {'name': activity_name, 'version': activity_version}
        }

        if input_arg is not None:
            attributes['input'] = utils.encode_task_input(input_arg)

        if timeout:
            attributes['startToCloseTimeout'] = str(timeout)

        if task_list:
            attributes['taskList'] = {'name': task_list}

        self.decisions.append({
            'decisionType': 'ScheduleActivityTask',
            'scheduleActivityTaskDecisionAttributes': attributes
        })

    def schedule_child_workflow_invocation(self, invocation_id, name, version, input_arg=None,
                                           child_policy=None, lambda_role=None, task_list=None,
                                           execution_start_to_close_timeout=None):
        attributes = {
            'workflowId': invocation_id,
            'workflowType': {'name': name, 'version': version}
        }

        if input_arg is not None:
            attributes['input'] = utils.encode_task_input(input_arg)

        if child_policy is not None:
            attributes['childPolicy'] = child_policy

        if lambda_role is None:
            lambda_role = self.workflow_state.lambda_role

        if lambda_role is not None:
            attributes['lambdaRole'] = lambda_role

        if task_list is not None:
            attributes['taskList'] = {'name': task_list}

        if execution_start_to_close_timeout is not None:
            attributes['executionStartToCloseTimeout'] = str(execution_start_to_close_timeout)

        self.decisions.append({
            'decisionType': 'StartChildWorkflowExecution',
            'startChildWorkflowExecutionDecisionAttributes': attributes
        })

    def start_timer(self, invocation_id, seconds):
        self.decisions.append({
            'decisionType': 'StartTimer',
            'startTimerDecisionAttributes': {
                'startToFireTimeout': str(seconds),
                'timerId': invocation_id
            }
        })

    def set_marker(self, invocation_id, value):
        self.decisions.append({
            'decisionType': 'RecordMarker',
            'recordMarkerDecisionAttributes': {
                'markerName': invocation_id,
                'details': utils.encode_task_input(value)
            }
        })

    def schedule_signal(self, invocation_id, workflow_id, input_arg=None,  run_id=None, control=None):
        attributes = {
            'signalName': invocation_id,
            'workflowId': workflow_id
        }

        if input_arg is not None:
            attributes['input'] = utils.encode_task_input(input_arg)

        if run_id is not None:
            attributes['runId'] = run_id

        if control is not None:
            attributes['control'] = control

        self.decisions.append({
            'decisionType': 'SignalExternalWorkflowExecution',
            'signalExternalWorkflowExecutionDecisionAttributes': attributes
        })

    def complete_workflow(self, result):
        """
        Schedule a CompleteWorkflow decision to indicate the workflow successfully completed

        :param result: The value to set as the result of the workflow
        """
        self.workflow_state.completed = True
        self.decisions.append({
            'decisionType': 'CompleteWorkflowExecution',
            'completeWorkflowExecutionDecisionAttributes': {
                'result': utils.encode_task_input(result)
            }
        })

    def fail_workflow(self, reason, details):
        """
        Schedule a FailWorkflowExecution decision to indicate the workflow failed

        :param reason: A short string describing why the workflow failed
        :param details: A longer description of the failure
        """
        self.workflow_state.completed = True
        self.decisions.append({
            'decisionType': 'FailWorkflowExecution',
            'failWorkflowExecutionDecisionAttributes': {
                'reason': reason,
                'details': details[:32768]
            }
        })

    def cancel_workflow(self, details):
        self.workflow_state.completed = True
        attributes = {}
        if details is not None:
            attributes['details'] = details
        self.decisions.append({
            'decisionType': 'CancelWorkflowExecution',
            'cancelWorkflowExecutionDecisionAttributes': attributes
        })

    def event_invocation_id(self, event):
        """
        Return the invocation id associated with an event
        :param event: SWF event object
        :return: The invocation id, or None if not found
        """
        attributes = self.root_event_attributes(event)
        candidate_keys = ['id', 'activityId', 'timerId', 'workflowId', 'markerName', 'signalName']
        for key in candidate_keys:
            if key in attributes:
                return attributes[key]

        return None

    @staticmethod
    def event_attributes(event):
        """
        Return the attributes dict associated with an event

        :param event: SWF event object
        :return: The attributes dict
        """
        event_type = event['eventType']
        attributes_key = '{}{}EventAttributes'.format(
            event_type[0].lower(), event_type[1:])
        return event.get(attributes_key)

    def root_event_attributes(self, event):
        """Return the attributes of the root event of an event.

        Many events concerning the progress of an activity refer back
        to an earlier event, rather than duplicate information from
        that earlier event.  For example, the ActivityTaskScheduled
        event contains a bunch of info about an activity that was
        scheduled.  Subsequent events concerning that activity just
        refer back to the ActivityTaskScheduledEvent rather than
        repeat the info.  This function finds the original event and
        returns its attributes.

        :param event: The event in question
        :return: The attributes of the root event, or attributes of this
          event if this event doesn't refer to another
          event.

        """
        event_attrs = self.event_attributes(event)
        if not event_attrs:
            return None

        root_id = event_attrs.get('scheduledEventId',
                                  event_attrs.get('initiatedEventId',
                                                  event_attrs.get('startedEventId')))
        if root_id is not None:
            found = [e for e in self.events if e['eventId'] == root_id]
            if found:
                return self.event_attributes(found[0])
            else:
                return event_attrs
        else:
            return event_attrs




from pyflow import exceptions
from pyflow import workflow_state as ws


class Future(object):
    """Abstract base class of Futures"""
    @property
    def done(self):
        raise NotImplementedError()

    @property
    def succeeded(self):
        raise NotImplementedError()

    @property
    def failed(self):
        raise NotImplementedError()

    @property
    def exception(self):
        raise NotImplementedError()

    def result(self):
        raise NotImplementedError()


class InvocationFuture(Future):
    """
    Represents the result of an asynchronous invocation.
    """

    def __init__(self, invocation, decision_helper):
        self._invocation = invocation
        self._decision_helper = decision_helper

    @property
    def done(self):
        """Return true if the invocation represented by this future is complete, successfully or not."""
        return self._invocation.done

    @property
    def succeeded(self):
        """Return true if the invocation represented by this future completed successfully."""
        return self._invocation.state == ws.InvocationState.SUCCEEDED

    @property
    def failed(self):
        """Return true if the invocation represented by this future completed with an exception being thrown."""
        return self.done and not self.succeeded

    @property
    def exception(self):
        """If this future is done, and failed, return the Exception thrown, else return None"""
        if self.failed:
            state = self._invocation.state
            if state == ws.InvocationState.TIMED_OUT:
                exception = exceptions.InvocationTimedOutException('InvocationStep {!r} timed out'.format(
                    self._invocation.invocation_id))
            elif state == ws.InvocationState.CANCELED:
                exception = exceptions.InvocationCanceledException(
                    self._invocation.failure_reason, self._invocation.failure_details)
            elif state == ws.InvocationState.FAILED:
                exception = exceptions.InvocationFailedException(
                    self._invocation.failure_reason, self._invocation.failure_details)
            else:
                exception = exceptions.DeciderException('Unexpected done state: {!r}'.format(state))
            return exception
        else:
            return None

    @property
    def invocation_id(self):
        """
        Return the id of the invocation this is waiting for.  Mainly useful for debugging and testing.
        """
        return self._invocation.invocation_id

    def result(self):
        """
        Blocks until either the invocation is complete, or runs out of events to process.  If invocation is complete,
        returns the result of the invocation if it succeeded, or raise exception if invocation failed.  If runs out
        events while waiting, raises WorkflowBlockedException.
        """
        while True:
            if self.done:
                if self.succeeded:
                    return self._invocation.result
                else:
                    raise self.exception

            if self._decision_helper.process_next_decision_task() is None:
                raise exceptions.WorkflowBlockedException()


class WrappedFuture(Future):
    """
    An instance of this class wraps another future and delegates all methods to the wrapped future.  Subclasses of
    this class can override any method to change its behavior.
    """

    def __init__(self, base_future):
        """
        Wrap another future.

        :param base_future: The Future to wrap
        """
        self._base = base_future

    @property
    def done(self):
        return self._base.done

    @property
    def failed(self):
        return self._base.failed

    @property
    def succeeded(self):
        return self._base.succeeded

    def result(self):
        return self._base.result()

    @property
    def exception(self):
        return self._base.exception



from pyflow import exceptions


class Workflow(object):
    """An instance of this class represents a workflow type.  Instances
    of this workflow can be executed.
    """

    # Workflow subclasses must override this
    NAME = None

    VERSION = '1.0'

    # default workflow options
    DEFAULT_OPTIONS = dict(
        defaultChildPolicy='TERMINATE',
        # default timeout in seconds for a task in this workflow to complete
        defaultTaskStartToCloseTimeout=600,
        # default timeout in seconds for an instance of this workflow to complete
        defaultExecutionStartToCloseTimeout=3600)

    # Workflow subclasses can redefine this to override default workflow options
    OPTIONS = {}

    def __init__(self, swf):
        if not (self.NAME and self.VERSION):
            raise exceptions.PyflowException('Workflow classes must define the NAME and VERSION properties')

        self._swf = swf
        ":type: pyflow.workflow_invocation_helper.WorkflowInvocationHelper"

    @property
    def name(self):
        """The name of this workflow"""
        return self.NAME

    @property
    def version(self):
        """The version string of this workflow"""
        return self.VERSION

    @property
    def options(self):
        """Additional options to pass to SWF when creating the workflow type"""
        options = self.DEFAULT_OPTIONS.copy()
        options.update(self.OPTIONS)
        return options

    @property
    def swf(self):
        return self._swf

    def run(self, workflow_input):
        """Subclasses override this method to implement the workflow behavior"""
        raise NotImplementedError()



import json
import logging
import uuid


# logger used for internal pyflow logging
logger = logging.getLogger('pyflow')


def encode_task_input(input_obj):
    return json.dumps(input_obj)


def decode_task_result(result_str):
    if result_str is None:
        return None
    else:
        return json.loads(result_str)


def iter_collection(paginator, items_key):
    for page in paginator:
        for item in page[items_key]:
            yield item


def list_domains(swf):
    paginator = swf.get_paginator('list_domains')
    return iter_collection(
        paginator.paginate(registrationStatus='REGISTERED'), 'domainInfos')


def list_workflow_types(swf, domain):
    paginator = swf.get_paginator('list_workflow_types')
    return iter_collection(
        paginator.paginate(domain=domain, registrationStatus='REGISTERED'),
        'typeInfos')


def list_activity_types(swf, domain):
    paginator = swf.get_paginator('list_activity_types')
    return iter_collection(
        paginator.paginate(domain=domain, registrationStatus='REGISTERED'),
        'typeInfos')


def list_workflow_history(swf, domain, workflow_id, run_id, reverse=False):
    paginator = swf.get_paginator('get_workflow_execution_history')
    return iter_collection(
        paginator.paginate(domain=domain, execution={'workflowId': workflow_id, 'runId': run_id}, reverseOrder=reverse),
        'events')


def poll_for_decision_tasks(swf, domain, task_list, identity):
    paginator = swf.get_paginator('poll_for_decision_task')
    result = None
    for page in paginator.paginate(domain=domain, taskList={'name': task_list},
                                   identity=identity):
        if result is None:
            result = page
        else:
            result['events'].extend(page['events'])
    return result


def poll_for_activity_tasks(swf, domain, task_list, identity):
    return swf.poll_for_activity_task(domain=domain,
                                      taskList={'name': task_list},
                                      identity=identity)


def schedule_activity_task(swf, task, task_list, activity_type, **options):
    attribs = {'activityType': activity_type,
               'activityId': 'activityId-' + str(uuid.uuid4()),
               'taskList': {'name': task_list}}
    attribs.update(options)

    swf.respond_decision_task_completed(
        taskToken=task['taskToken'],
        decisions=[
            {
                'decisionType': 'ScheduleActivityTask',
                'scheduleActivityTaskDecisionAttributes': attribs
            }
        ])


def init_domain(swf):
    """
    :param swf: boto3 SWF client
    :return: domain name for this project
    """
    domain_name = 'SWFSampleDomain'

    # see if domain already exists.  If not, register it.
    for d in list_domains(swf):
        if d['name'] == domain_name:
            break
    else:
        swf.register_domain(name=domain_name,
                            description='{} domain'.format(domain_name),
                            workflowExecutionRetentionPeriodInDays='1')
    return domain_name



import functools

import attr

from pyflow import exceptions

__all__ = ['LambdaDescriptor', 'ActivityDescriptor', 'ChildWorkflowDescriptor']


@attr.s
class LambdaDescriptor(object):
    """
    Used to define a shortcut for invoking a lambda function in a workflow definition.

        LambdaDescriptor(function_name, timeout=None)

    Usage::

        class MyWorkflow(pyflow.Workflow):
            # The lambda function name is a required parameter
            my_function = pyflow.LambdaDescriptor('my_function')
            # Can also optionally specify a timeout
            other_function = pyflow.LambdaDescriptor('other_lambda_func', timeout=120)

            def run(self, input):
                # these two calls are equivalent
                fut1 = self.my_function(input)
                fut2 = self.swf.invoke_lambda('my_function', input)

                # these calls are equivalent
                fut3 = self.other_function(fut2.result())
                fut4 = self.invoke_lambda('other_lambda_func', fut2.result(), timeout=120)
    """
    function_name = attr.ib()
    timeout = attr.ib(default=None)

    def __get__(self, instance, owner):
        return functools.partial(self._invoke, instance)

    def _invoke(self, instance, *args, **kwargs):
        if (args and kwargs) or len(args) > 1:
            raise exceptions.InvocationException('Lambda must be invoked with either a single positional argument, '
                                                 'or any number of keyword arguments, but not both.')

        if args:
            input_arg = args[0]
        elif kwargs:
            input_arg = kwargs
        else:
            input_arg = None

        return instance.swf.invoke_lambda(self.function_name, input_arg, timeout=self.timeout)


@attr.s
class ActivityDescriptor(object):
    """
    Used to define a shortcut for calling an SWF activity.

        ActivityDescriptor(activity_name, activity_version, timeout=None, task_list=None)

    Usage::

        class MyWorkflow(pyflow.Workflow):
            activity1 = pyflow.ActivityDescriptor('activity_name1', 'v1.0')
            activity2 = pyflow.ActivityDescriptor('activity_name2', 'v1.0', timeout=120, task_list='some-task-list')

            def run(self, input):
                # these calls are equivalent
                fut1 = self.activity1(input)
                fut2 = self.invoke_activity('activity_name1', 'v1.0', input)

                # these are equivalent
                self.activity2(input)
                self.invoke_activity('activity_name2', 'v1.0', input,
                    timeout=120, task_list='some-task-list')
    """
    activity_name = attr.ib()
    activity_version = attr.ib()
    timeout = attr.ib(default=None)
    task_list = attr.ib(default=None)

    def __get__(self, instance, owner):
        return functools.partial(self._invoke, instance)

    def _invoke(self, instance, *args, **kwargs):
        if (args and kwargs) or len(args) > 1:
            raise exceptions.InvocationException('Activity must be invoked with either a single positional argument, '
                                                 'or any number of keyword arguments, but not both.')

        if args:
            input_arg = args[0]
        elif kwargs:
            input_arg = kwargs
        else:
            input_arg = None

        return instance.swf.invoke_activity(self.activity_name, self.activity_version, input_arg=input_arg,
                                            timeout=self.timeout, task_list=self.task_list)


@attr.s
class ChildWorkflowDescriptor(object):
    """
    Used to define a shortcut for invoking a child workflow.

        ChildWorkflowDescriptor(workflow_name, workflow_version, child_policy=None, lambda_role=None,
            task_list=None, execution_start_to_close_timeout=None)

    Usage::

        class MyWorkflow(pyflow.Workflow):
            some_workflow = pyflow.ChildWorkflowDescriptor('SomeWorkflow', '1.0', task_list='some_task_list')

            def run(self, input)
                # these are equivalent
                fut = self.some_workflow(input)
                fut2 = self.swf.invoke_child_workflow('SomeWorkflow', '1.0', input, task_list='some_task_list'
    """
    workflow_name = attr.ib()
    workflow_version = attr.ib()
    child_policy = attr.ib(default=None)
    lambda_role = attr.ib(default=None)
    task_list = attr.ib(default=None)
    execution_start_to_close_timeout = attr.ib(default=None)

    def __get__(self, instance, owner):
        return functools.partial(self._invoke, instance)

    def _invoke(self, instance, *args, **kwargs):
        if (args and kwargs) or len(args) > 1:
            raise exceptions.InvocationException('Activity must be invoked with either a single positional argument, '
                                                 'or any number of keyword arguments, but not both.')

        if args:
            input_arg = args[0]
        elif kwargs:
            input_arg = kwargs
        else:
            input_arg = None

        return instance.swf.invoke_child_workflow(
            self.workflow_name, self.workflow_version, input_arg=input_arg,
            child_policy=self.child_policy, lambda_role=self.lambda_role,
            task_list=self.task_list,
            execution_start_to_close_timeout=self.execution_start_to_close_timeout)



import traceback

import dateutil.tz

from pyflow import exceptions
from pyflow import future
from pyflow import workflow_state as ws


class InvokeOnceFuture(future.WrappedFuture):
    """A Future subclass returned by invoke_once"""
    def result(self):
        base_result = super(InvokeOnceFuture, self).result()
        if base_result['succeeded']:
            return base_result['result']
        else:
            raise exceptions.InvocationFailedException(base_result['reason'], base_result['details'])


class WorkflowInvocationHelper(object):
    # Possible arguments for the child_policy parameter of invoke_child_workflow
    CHILD_POLICY_TERMINATE = 'TERMINATE'
    CHILD_POLICY_REQUEST_CANCEL = 'REQUEST_CANCEL'
    CHILD_POLICY_ABANDON = 'ABANDON'

    """
    Provides the interface and functionality that workflow functions use to interact with the decider.
    """

    def __init__(self, decision_helper):
        self._workflow_state = decision_helper.workflow_state
        ":type: pyflow.workflow_state.WorkflowState"

        self._decision_helper = decision_helper
        ":type: pyflow.decision_task_helper.DecisionTaskHelper"

        self._invocation_counters = {}

    @property
    def run_id(self):
        """The runId of the currently executing workflow"""
        return self._workflow_state.run_id

    @property
    def workflow_id(self):
        """The workflowId of the currently running workflow"""
        return self._workflow_state.workflow_id

    @property
    def lambda_role(self):
        """The lambdaRole of the currently running workflow"""
        return self._workflow_state.lambda_role

    @property
    def is_replaying(self):
        """True if the current statement is being replayed"""
        return self._decision_helper.is_replaying

    def workflow_time(self):
        """Return the current time according to the workflow clock"""
        return self._decision_helper.last_event_timestamp.astimezone(dateutil.tz.tzutc())

    def invoke_lambda(self, function_name, input_arg, timeout=None):
        """
        Schedule a lambda invocation

        :param function_name: The name of the lambda function to invoke
        :param input_arg: Input argument for the lambda.  Must be a JSON-serializable object.
        :param timeout: Timeout value, in seconds, after which the lambda function is considered to have failed
          if it has not returned.  It can be any integer from 1-300.  If not specified, defaults to 300s.
        :return: A Future which yields the result of the lambda function
        """
        invocation_id = self._next_invocation_id('lambda')
        invocation_state = self._workflow_state.get_invocation_state(invocation_id)
        out_fut = future.InvocationFuture(invocation_state, self._decision_helper)

        if invocation_state.state == ws.InvocationState.NOT_STARTED:
            invocation_state.update_state(ws.InvocationState.HANDLED)

            if not self._decision_helper.is_replaying:
                self._decision_helper.schedule_lambda_invocation(invocation_id, function_name, input_arg, timeout)

        return out_fut

    def invoke_activity(self, name, version, input_arg=None, timeout=None, task_list=None):
        """
        Schedule a SWF activity invocation

        :param name: Name of the activity type
        :param version: Version of the activity type
        :param input_arg: Input argument for the activity.  Must be a JSON-serializable object.
        :param timeout: Timeout value, in seconds, after which the activity is considered to have failed
          if it has not returned.
        :param task_list: Name of the SWF task list that the activity worker is listening for events on.  If not given,
          The default registered with the activity type will be used.
        :return: A Future which yields the result of the activity.
        """
        invocation_id = self._next_invocation_id('activity')
        invocation_state = self._workflow_state.get_invocation_state(invocation_id)
        out_fut = future.InvocationFuture(invocation_state, self._decision_helper)

        if invocation_state.state == ws.InvocationState.NOT_STARTED:
            invocation_state.update_state(ws.InvocationState.HANDLED)

            if not self._decision_helper.is_replaying:
                self._decision_helper.schedule_activity_invocation(invocation_id, name, version, input_arg,
                                                                   timeout=timeout, task_list=task_list)

        return out_fut

    def invoke_child_workflow(self, name, version, input_arg=None, child_policy=None, lambda_role=None, task_list=None,
                              execution_start_to_close_timeout=None):
        """
        Schedule a child workflow invocation.

        For all of the optional arguments, if they aren't given then the defaults specified when registering the child
        workflow type will be used.

        :param name: The name of the child workflow type
        :param version: Version of child workflow type
        :param input_arg: Input argument to the workflow.  Must be a JSON-serializable object
        :param child_policy: Specify the policy to use for the child workflow execution if this workflow is terminated
          or times out.  Can be one of the class constants CHILD_POLICY_TERMINATE, CHILD_POLICY_REQUEST_CANCEL, or
          CHILD_POLICY_ABANDON.
        :param lambda_role: The IAM role the child workflow will use to invoke lambdas as.
        :param task_list: Name of the SWF task list that the decider for the child workflow is listening for events on.
        :param execution_start_to_close_timeout: Total duration for this workflow execution in seconds, after which it
          will be considered to have timed out if it hasn't finished.
        :return: A Future which yields the result of this child workflow
        """
        invocation_id = self._next_invocation_id('child_workflow')
        invocation_state = self._workflow_state.get_invocation_state(invocation_id)
        out_fut = future.InvocationFuture(invocation_state, self._decision_helper)

        if invocation_state.state == ws.InvocationState.NOT_STARTED:
            invocation_state.update_state(ws.InvocationState.HANDLED)

            if not self._decision_helper.is_replaying:
                self._decision_helper.schedule_child_workflow_invocation(
                    invocation_id, name, version, input_arg=input_arg, child_policy=child_policy,
                    lambda_role=lambda_role, task_list=task_list,
                    execution_start_to_close_timeout=execution_start_to_close_timeout)

        return out_fut

    def invoke_once(self, callable, *args, **kwargs):
        """
        Invoke a python callable asynchronously, ensuring it is only invoked once in a workflow invocation.

        The result of the invocation will be propagated across replays of the workflow.  Using invoke_once provides
        similar behavior to if you turned callable into a lambda and invoked it with invoke_lambda, except that the
        callable is executed locally in the decider.  This can be useful because the function is too
        trivial to be worth deploying as a lambda, or because you want to pass input arguments to it that can't be
        serialized as JSON.

        :param callable: A python callable
        :param args: positional arguments to pass to callable
        :param kwargs: keyword arguments to pass to callable
        :return: A Future which yields the result of invoking callable.  Both error and normal return values will be
          preserved.
        """
        invocation_id = self._next_invocation_id('invoke_once')
        signal_id = self._next_invocation_id('invoke_once_signal')
        invocation_state = self._workflow_state.get_invocation_state(invocation_id)
        marker_fut = future.InvocationFuture(invocation_state, self._decision_helper)
        out_fut = InvokeOnceFuture(marker_fut)

        if invocation_state.state == ws.InvocationState.NOT_STARTED:
            invocation_state.update_state(ws.InvocationState.HANDLED)

            if not self._decision_helper.is_replaying:
                try:
                    result = callable(*args, **kwargs)
                except Exception:
                    marker_val = {'succeeded': False,
                                  'reason': 'Exception raised while invoking callable',
                                  'details': traceback.format_exc()}
                else:
                    marker_val = {'succeeded': True, 'result': result}

                self._decision_helper.set_marker(invocation_id, marker_val)
                self._decision_helper.schedule_signal(signal_id, self._decision_helper.workflow_id,
                                                      run_id=self._decision_helper.run_id)

        return out_fut

    def start_timer(self, seconds):
        """Returns a Future which will be done in the given number of seconds.

        :param seconds: Number of seconds until timer will be done
        :return: A Future which yields None when the timer is done.
        """
        invocation_id = self._next_invocation_id('sleep')
        invocation_state = self._workflow_state.get_invocation_state(invocation_id)
        out_fut = future.InvocationFuture(invocation_state, self._decision_helper)

        if invocation_state.state == ws.InvocationState.NOT_STARTED:
            invocation_state.update_state(ws.InvocationState.HANDLED)

            if not self._decision_helper.is_replaying:
                self._decision_helper.start_timer(invocation_id, seconds)

        return out_fut

    def sleep(self, seconds):
        """
        Blocks for the specified amount of time.

        :param seconds: Number of seconds to sleep
        """
        return self.start_timer(seconds).result()

    def wait_for_all(self, *futures):
        """
        Waits for all the futures to be done.

        :param futures: One or more futures.  If any arguments are a list, they are assumed to be a list of futures.
        :return: A list containing the input futures, in the order that they finished.
        """
        return self._wait_for_condition(lambda fs: all(f.done for f in fs), *futures)

    def wait_for_any(self, *futures):
        """
        Waits for any of the futures to be done.

        :param futures: One or more futures.  If any arguments are a list, they are assumed to be a list of futures.
        :return: A list containing the futures that finished, in the order that they finished.
        """
        return self._wait_for_condition(lambda fs: any(f.done for f in fs), *futures)

    def timed_wait_for_all(self, timeout, *futures):
        """
        Wait for all futures to be done, with a timeout.

        :param timeout: Timeout in seconds to wait
        :param futures: One or more futures.  If any arguments are a list, they are assumed to be a list of futures.
        :return: A list containing the futures that finished, in the order that they finished.
        :raises WaitTimedOutException: If all futures are not done within timeout seconds.
        """
        return self._timed_wait_for_condition(lambda fs: all(f.done for f in fs), timeout, *futures)

    def timed_wait_for_any(self, timeout, *futures):
        """
        Wait for any futures to be done, with a timeout.

        :param timeout: Timeout in seconds to wait
        :param futures: One or more futures.  If any arguments are a list, they are assumed to be a list of futures.
        :return: A list containing the futures that finished, in the order that they finished.
        :raises WaitTimedOutException: If no futures are done within timeout seconds.
        """
        return self._timed_wait_for_condition(lambda fs: any(f.done for f in fs), timeout, *futures)

    def _wait_for_condition(self, predicate, *futures):
        """
        Waits for a condition to become true on a list of futures.  This is a helper for wait_for_all and
        wait_for_any.

        :param predicate: A callable which will be called with the input list of futures before each round of
          event processing.  It should return true when the condition is satisfied which should cause the function to
          return.
        :param futures: One or more futures.  If any arguments are a list, they are assumed to be a list of futures.
        :return: A list of the input futures which were done when the predicate returned True, in the order that they
          finished.
        """
        if any(isinstance(f, list) for f in futures):
            flattened = []
            for f in futures:
                if isinstance(f, list):
                    flattened.extend(f)
                else:
                    flattened.append(f)
            futures = flattened

        not_done = list(futures)
        done = []
        while True:
            done.extend(f for f in not_done if f.done)
            not_done = [f for f in not_done if not f.done]
            if predicate(futures):
                return done
            elif self._decision_helper.process_next_decision_task() is None:
                raise exceptions.WorkflowBlockedException()

    def _timed_wait_for_condition(self, predicate, timeout, *futures):
        """
        Waits for a condition to become true on a list of predicates, with a timeout.

        :param predicate: Same meaning as for _wait_for_condition
        :param timeout: Timeout in seconds to wait
        :param futures: One or more futures.  If any arguments are a list, they are assumed to be a list of futures.
        :return: A list containing the futures that finished, in the order that they finished.
        :raises WaitTimedOutException: If no futures are done within timeout seconds.
        """
        timeout_fut = self.start_timer(timeout)

        done_futures = self._wait_for_condition(lambda futs: predicate(futs) or timeout_fut.done,
                                                *futures)
        if predicate(futures):
            return done_futures
        else:
            raise exceptions.WaitTimedOutException()

    def _next_invocation_id(self, prefix):
        """
        Deterministically generates the next invocation id for a given prefix.  The id will be unique in the context
        of this DecisionTaskHelper instance.

        :param prefix: Prefix of the invocation id
        :return: A new invocation id
        """
        new_id = self._invocation_counters.get(prefix, 0) + 1
        self._invocation_counters[prefix] = new_id
        return '{}{}'.format(prefix, new_id)




class PyflowException(Exception):
    """Base class of exceptions raised by Pyflow"""
    pass


class DeciderException(PyflowException):
    """Exception raised for errors in the decider"""
    pass


# This class is intentionally not derived from PyflowException, so that workflow functions can catch PyflowException
# for error handling without interfering with pyflow.
class WorkflowBlockedException(Exception):
    """Exception thrown from a wait_for* method to exit to the event loop when a workflow execution is blocked."""
    pass


class WorkflowFailedException(PyflowException):
    """
    Exception which can be thrown by a workflow function to explicitly request that the workflow be marked as failed,
    with a given reason and details string.
    """
    def __init__(self, reason, details):
        super(WorkflowFailedException, self).__init__(reason, details)

    @property
    def reason(self):
        return self.args[0]

    @property
    def details(self):
        return self.args[1]


class InvocationException(PyflowException):
    """Base class of exceptions related to invocations failing."""
    pass


class InvocationTimedOutException(InvocationException):
    """Exception thrown to indicate a timeout occurred"""
    pass


class InvocationFailedException(InvocationException):
    """Exception thrown to indicate an invocation failed"""
    def __init__(self, reason=None, details=None):
        msg = reason or ''

        if details:
            if msg:
                msg += ': ' + details
            else:
                msg = details

        super(InvocationFailedException, self).__init__(msg)


class InvocationCanceledException(InvocationException):
    """Exception thrown to indicate an invocation was canceled"""

    def __init__(self, reason=None, details=None):
        msg = reason or ''

        if details:
            if msg:
                msg += ': ' + details
            else:
                msg = details

        super(InvocationFailedException, self).__init__(msg)


class WaitTimedOutException(PyflowException):
    """Exception thrown to indicate a timeout occurred while waiting for a result"""
    pass



import os

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    long_description = f.read()

setup(
    name='pyflow-swf',
    version='1.5.0',
    author='Adam Jenkins',
    author_email='adam@thejenkins.org',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2.7',
    ],
    packages=['pyflow'],
    url='https://github.com/ajenkins-cargometrics/pyflow',
    description='Python Workflow Library built on the AWS SWF service',
    long_description=long_description,
    install_requires=[
        'boto3',
        'attrs',
        'python-dateutil'
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest>=3.0',
    ]
)
